% Written by Amit Hochman.
% Released under the Gnu Public License.
% Last updated on: 6/25/2009.

function [oSdpTh, oSdpDn, oBcD, oBcU, oSc, oIm, oPole, oMasks] = calc_contours(vX, vY, x_s, y_s, k0, Er_d, Er_u, points, sComp, isMSource, qsRule)
bc2sdpPoints = 0.50;
sc_factor = 2;
bc_points = points*bc2sdpPoints;
kc_d = sqrt(Er_d)*k0;
kc_u = sqrt(Er_u)*k0;
kc_th = min(kc_d, kc_u);
kc_dn = max(kc_d, kc_u);
delta_k = kc_dn-kc_th;
Er_dn = max(Er_d, Er_u);
Er_th = min(Er_d, Er_u);
vX = vX(:);
vY = vY(:);
nX = length(vX);
bDid = Er_d>Er_u;
bvRef = vY<=0;
bvTran = ~bvRef;
vA = abs(vX-x_s);
vB = bvTran.*vY;
vC = (abs(y_s)+bvRef.*abs(vY));
vR = hypot(vA, vB + vC);
bvXgrY = (vA > vB + vC);
bvXggY = (vA(bvTran) > 10*(vB(bvTran) + vC(bvTran))); % sdp calculation doesn't work well when x>>y
bvSmall_contrast = (delta_k*vA<1.5);

vB_r = vB(bvRef);
vB_t = vB(bvTran);
vC_r = vC(bvRef);
vC_t = vC(bvTran);

oIm.Il = zeros(nX,1);
oIm.Er = zeros(nX,1);
oIm.bQs = zeros(nX,1);
oIm.y_s = zeros(nX,1);
oIm.x_s = repmat(x_s, nX, 1);
oSdpTh = [];
oSdpDn = [];
oBcD = [];
oBcU = [];
oSc = [];

% calc. asymptotic slopes
vM1 = vA./(vB + vC);
bvSing = vB == vC;
vM2(bvSing) = inf;
vM2(~bvSing) = bvTran(~bvSing).*(vA(~bvSing)./(vC(~bvSing)-vB(~bvSing)).*...
    (2*bDid-1)) + bvRef(~bvSing).*vM1(~bvSing)*(2*bDid-1);
vM2 = vM2(:);
vM3 = -vM1;
vA2 = atan(vM2);
vA2 = vA2.*(vA2>0) + (vA2+pi).*(vA2<0);
% calc. BCI angles
vTheta_dn = (vA2+atan(vM3)-pi)/2;
vTheta_th = (atan(vM1)+vA2)/2+pi;
vStep_th = exp(1j*vTheta_th);
vStep_dn = exp(1j*vTheta_dn);
vStep_down = bDid*vStep_dn+~bDid*vStep_th;
vStep_up = ~bDid*vStep_dn+bDid*vStep_th;

dense = true;
thin = false;
vAlpha_th = calc_alpha(thin, vStep_th);
vAlpha_dn = calc_alpha(dense, vStep_dn);
vAlpha_down = bDid*vAlpha_dn+~bDid*vAlpha_th;
vAlpha_up = ~bDid*vAlpha_dn+bDid*vAlpha_th;

vAngleForSc = -max(abs(atan(vM3)),pi/4);
vStep_right_sdp = exp(1j*vAngleForSc);
vAlpha_sdp_right = -vA.*abs(imag(vStep_right_sdp)) - (vB+vC).*abs(real(vStep_right_sdp));
vAlpha_sdp = vAlpha_sdp_right;

vPowD(bvTran) = 0;
vPowU(bvTran) = 0.5*(lower(sComp(2))=='x');
vPowU(bvRef) = 0;
vPowD(bvRef) = -0.5*((lower(sComp(2))=='y') || (lower(sComp(2))=='z'));

% the difference integrand has a sqrt dependence.
vPowD = vPowD + (vPowD==0)*0.5;
vPowU = vPowU + (vPowU==0)*0.5;
vPowD = vPowD(:);
vPowU = vPowU(:);
vPowDn = bDid*vPowD + ~bDid*vPowU;
vPowTh = ~bDid*vPowD + bDid*vPowU;

% High contrast, MLS => leaky pole
k_pole = sqrt((Er_d*Er_u)/(Er_d+Er_u))*k0;
delta_k_pole = abs(k_pole-kc_th);
bvHigh_contrast_th = (delta_k_pole*vA<1);

mSdp = sdp(vX, vY, x_s, y_s, kc_d, kc_u);
sdp_r = mSdp(bvRef,1);
sdp_th_t = mSdp(bvTran,1);
sdp_dn_t = mSdp(bvTran,2);
sigma_r = sqrt(1./abs((second_deriv(sdp_r, vB_r, vC_r, kc_d, kc_u))))*sqrt(2);
sigma_th_t = sqrt(1./abs((second_deriv(sdp_th_t, vB_t, vC_t, kc_d, kc_u))))*sqrt(2);
vAlpha2 = second_deriv(sdp_dn_t, vB_t, vC_t, kc_d, kc_u);
vGamma_CplxSp = pi/4*ones(nX,1);
vGamma_CplxSp(bvTran) = 1/2*angle(conj(1j*vAlpha2));
vGamma_CplxSp(vGamma_CplxSp<0) = vGamma_CplxSp(vGamma_CplxSp<0)+pi;
sigma_dn_t = sqrt(1./abs(vAlpha2))*sqrt(2);

vSigma_th = zeros(nX, 1);
vSigma_dn = zeros(nX, 1);
vSigma_th(bvTran) = sigma_th_t;
vSigma_th(bvRef) = sigma_r;
vSigma_dn(bvTran) = sigma_dn_t;
vSigma_dn(bvRef) = sigma_r;

bvQs = 2*kc_dn < 2*pi./vR;
oIm.bQs = bvQs;

vSdp_th = zeros(nX, 1);
vSdp_dn = zeros(nX, 1);
vSdp_th(bvTran) = sdp_th_t;
vSdp_th(bvRef) = sdp_r;
vSdp_dn(bvTran) = sdp_dn_t;
vSdp_dn(bvRef) = sdp_r;

% When SP near BP, the SDP is moved so that it passes right through the BP.
% The singularity is dealt with by an appropriate Gaussian rule.
vFalse_all = false(nX, 1);
vFalse_t = false(sum(bvTran),1);
vFalse_r = false(sum(bvRef),1);
bvCritJump = vFalse_all;
sigmas_crit = 1;
if any(bvRef)
    bvCritJump_r = (abs(sdp_r-kc_th) < sigma_r/sqrt(2)*sigmas_crit) & ~bvQs(bvRef) & bDid;
    warning('off', 'MATLAB:divideByZero');
    vSecondIntrsct = kc_d^2./sdp_r;
    warning('on', 'MATLAB:divideByZero');
    bvCritJump_r = bvCritJump_r & (abs(vSecondIntrsct-kc_dn) > (kc_dn-kc_th)/2 + abs(sdp_r-kc_th));
    bvCritJump(bvRef) = bvCritJump_r;
end

% Case B
% transmitted
sigmas_caseB = 1;
if any(bvTran)
    bvCaseBTth = abs(vSdp_th(bvTran)-kc_th) < vSigma_th(bvTran)/sqrt(2)*sigmas_caseB;
    bvCaseBTdn = abs(vSdp_dn(bvTran)-kc_dn) < vSigma_dn(bvTran)/sqrt(2)*sigmas_caseB;
    bvCaseBTth = bvCaseBTth | bvXggY;
    bvCaseBTdn = bvCaseBTdn | bvXggY;
end
% reflected
if any(bvRef)
    bvCaseBR = abs(sdp_r-kc_d) < sigma_r/sqrt(2)*sigmas_caseB;
end

% calc. whether dense SP is intercepted (in Trans.)
bvSpDnIntcp = vFalse_t;
oMasks.bvBpSpIntcp = vFalse_all;
if any(bvTran)
    isImpTh = 0;
    isImpDn = 0;
    s = vSdp_th(bvTran);
    ph_at_spTh = phase_factor(s, isImpTh, isImpDn, bvTran);
    isImpDn = 1;
    s = vSdp_dn(bvTran);
    ph_at_spDn = phase_factor(s, isImpTh, isImpDn, bvTran);
    bvSpDnIntcp = real(ph_at_spTh) < real(ph_at_spDn);
    bvSpDnIntcp = bvSpDnIntcp | bvXggY;
    oMasks.bvBpSpIntcp(bvTran) = bvSpDnIntcp;
end

% calc. whether Up BP is intercepted (in Ref.)
bvUpBpIntcp = vFalse_r;
if any(bvRef)
    bvUpBpIntcp = vA(bvRef)./sqrt(vA(bvRef).^2+vC(bvRef).^2)>sqrt(Er_th/Er_dn);
    oMasks.bvBpSpIntcp(bvRef) = bvUpBpIntcp;
    bvUpBpIntcp = bvUpBpIntcp & ~bvCritJump_r;
end

oPole.vStep_u = vStep_up;
oPole.vStep_d = vStep_down;
oPole.vGamma_u  = -vAlpha_up*kc_u;
oPole.vGamma_d  = -vAlpha_down*kc_d;
oPole.k_pole = k_pole;
oPole.vSing_d = vPowD;
oPole.vSing_u = vPowU;
% Control Section --------------------------------
% Quasi-static
if any(bvQs)
    add_qs;
end
bvDone = bvQs;
bvSDPth = ~bvDone;
% reflected
if any(bvRef)
    bvBCIRu = ~bvDone & bvRef;
    bvBCIRu(bvRef) = bvBCIRu(bvRef) & bvUpBpIntcp;
    bvBCIRd = ~bvDone & bvRef;
    bvBCIRd(bvRef) = bvBCIRd(bvRef) & bvUpBpIntcp & bvCaseBR;
    bvSDPth(bvRef) = bvSDPth(bvRef) & (~bvUpBpIntcp | bvUpBpIntcp & ~bvCaseBR);
end

% transmitted
if any(bvTran)
    bvSDPth(bvTran) = bvSDPth(bvTran) & (~bvSpDnIntcp | bvSpDnIntcp & ~bvCaseBTth);
    bvSDPdn = ~bvDone & bvTran;
    bvSDPdn(bvTran) = bvSDPdn(bvTran) & bvSpDnIntcp & ~bvCaseBTdn;
    bvBCIth = ~bvDone;
    bvBCIth(bvTran) = bvBCIth(bvTran) & bvSpDnIntcp  & bvCaseBTth;
    bvBCIdn = ~bvDone;
    bvBCIdn(bvTran) = bvBCIdn(bvTran) & bvSpDnIntcp  & bvCaseBTdn;
end

if any(bvRef)
    if bDid
        bvBCIdn(bvRef) = bvBCIRd(bvRef);
        bvBCIth(bvRef) = bvBCIRu(bvRef);
    else
        bvBCIth(bvRef) = bvBCIRd(bvRef);
        bvBCIdn(bvRef) = bvBCIRu(bvRef);
    end
end
bvBCIdn = bvBCIdn(:);
bvBCIth = bvBCIth(:);
% Small contrast
bvSc = bvSmall_contrast & ~bvQs;
if any(bvTran)
    bvSc(bvTran) = bvSc(bvTran) & (bvBCIdn(bvTran) | bvCaseBTth); 
end
if any(bvRef)
    bvScCrit = ((kc_dn-kc_th) <sigma_r) & (abs(sdp_r-kc_th) < sigma_r/sqrt(2)*sigmas_crit);
    bvSc(bvRef) = bvSc(bvRef) & bvScCrit;
end

bvSDPth = bvSDPth & ~bvSc;
if any(bvTran)
    bvSDPdn = bvSDPdn & ~bvSc;
end
bvBCIth = bvBCIth & ~bvSc;
bvBCIdn = bvBCIdn & ~bvSc;
if any(bvSc)
    add_sc;
end

% masks
if any(bvRef)
    oMasks.bvBCIRu = bvBCIRu;
    oMasks.bvBCIRd = bvBCIRd;
    oMasks.bvCritJump = bvCritJump;
end
if any(bvTran)
    oMasks.bvSDPdn = bvSDPdn;
end
oMasks.bvSDPth = bvSDPth;
oMasks.bvBCIth = bvBCIth;
oMasks.bvBCIdn = bvBCIdn;
oMasks.bvQs = bvQs;
oMasks.bvSc = bvSc;

if any(bvSDPth)
    bvFix = add_sdp(bvSDPth, thin);
else
    bvFix = false;
end

% fix possible error in dense bp interception due to inaccurate path.
% it is possible that the path will not intcp, when the exact path
% should intcp. This is detected by two conditions:
% 1. That the exact path intcps
% 2. That the path ended on the proper sheet.
% if bvFix, the path surrounding the Dense BP is unnecessary.
if any(bvTran)
    if any(bvSDPdn)
        if any(bvFix)
            bvSDPdn = bvSDPdn & ~bvFix;
        end
        if any(bvSDPdn)
            add_sdp(bvSDPdn, dense);
        end
    end
end
if any(bvBCIdn)
    if any(bvFix)
        bvBCIdn = bvBCIdn & ~bvFix;
    end
    if any(bvBCIdn)
        add_bci(bvBCIdn(:), dense);
    end
end
if any(bvBCIth)
    add_bci(bvBCIth(:), thin);
end
bvActive_dn = vFalse_all; 
bvCaseA = bDid & bvBCIth & bvRef & ~(bvHigh_contrast_th & isMSource);
bvActive_th = isMSource & ~bvCaseA;
oPole.vActive_d = bDid & bvActive_dn | ~bDid & bvActive_th;
oPole.vActive_u = bDid & bvActive_th | ~bDid & bvActive_dn;


% Clean-up -----------------------------------------------------------
if ~isempty(oBcD)
    % Squeeze out zeros
    % We do not know which observation points will require bci's before the
    % calculation has ended (depends on which RS we end up on)
    oBcD{1} = squeeze_zeros(oBcD{1});
    oBcD{2} = squeeze_zeros(oBcD{2});
end
if ~isempty(oBcU)
    oBcU{1} = squeeze_zeros(oBcU{1});
    oBcU{2} = squeeze_zeros(oBcU{2});
end
if ~isempty(oSdpTh)
    oSdpTh = squeeze_zeros(oSdpTh);
end
if ~isempty(oSdpDn)
    oSdpDn = squeeze_zeros(oSdpDn);
end
if ~isempty(oSc)
    oSc = squeeze_zeros(oSc);
end
% add_bci ---------------------------------------------------
    function add_bci(bv, bDense)
        bUp = bDid & ~bDense | ~bDid & bDense;
        bvCaseA = bUp & bv & bvRef & ~(bvHigh_contrast_th & isMSource);
        bvCaseAL = bvCaseA(bv); % local version
        if bDense
            vSing = vPowDn;
        else
            vSing = vPowTh;
        end
        vSing = vSing(bv);
        if bDense
            vAlpha = vAlpha_dn(bv);
            vStep = vStep_dn(bv);
        else
            vAlpha = vAlpha_th(bv);
            vStep = vStep_th(bv);
        end
        bvSingMin05 = vSing == -0.5;
        bvSingPls05 = vSing == 0.5;
        xiR = zeros(sum(bv), bc_points); % These are the actual sampling points
        w = xiR;                         % Not the ones used to track the path (in case a)
        kx = xiR;
        jac = xiR;
        if any(bvSingMin05)
            [xiR(bvSingMin05,:), w(bvSingMin05,:)] = gausslag(bc_points, abs(vAlpha(bvSingMin05,:)), -0.5);
        end
        if any(bvSingPls05)
            [xiR(bvSingPls05,:), w(bvSingPls05,:)] = gausslag(bc_points, abs(vAlpha(bvSingPls05,:)), 0.5);
        end
        kc = kc_th*~bDense+kc_dn*bDense;
        if any(~bvCaseAL) % straight-line path
            mStep = makematC(vStep(~bvCaseAL), bc_points);
            kx(~bvCaseAL,:) = mStep.*xiR(~bvCaseAL,:)+kc;
            jac(~bvCaseAL,:) = mStep.*w(~bvCaseAL,:);
        end
        if any(bvCaseAL)
            nCaseA = sum(bvCaseAL);
            nTrack = bc_points; % must be even for smooth_path
            TrackLen = 1.001;
            vMin = min(xiR(bvCaseAL,:), [], 2)*TrackLen;
            vMax = max(xiR(bvCaseAL,:), [], 2)*TrackLen;
            xi = mlinspace(vMin, vMax, nTrack);
            vGamma = angle(1j*vA(bvCaseAL)-1j*vC(bvCaseAL)*kc_u/sqrt(kc_d^2-kc_u^2));
            vGamma(vGamma < 0) = vGamma(vGamma<0)+pi;
            g0 = exp(1j*vGamma);
            isImpTh = ones(nCaseA, 1)*bDid;
            isImpDn = isImpTh;
            mKxCaseA = grad_walk(kc, g0, xi, isImpTh, isImpDn, bvCaseA);
            x = real(mKxCaseA);
            y = imag(mKxCaseA);
            iCaseAL = find(bvCaseAL);
            iCaseAL = iCaseAL(1);
            [kx(bvCaseAL,:), jac(bvCaseAL,:)] = smooth_path_vect_LS(x, y, xi(1,:), xiR(iCaseAL,:), vMax, repmat(kc,nCaseA,1));
            jac(bvCaseAL,:) = jac(bvCaseAL,:).*w(bvCaseAL,:);
        end
        bmImLineCross = real(kx) < 0;
        if bDense
            isImpDnR = bmImLineCross;
            isImpThR = bmImLineCross;
            isImpDnL = ~bmImLineCross;
            isImpThL = bmImLineCross;
        else
            isImpDnR = ~bmImLineCross;
            isImpThR = bmImLineCross;
            isImpDnL = ~bmImLineCross;
            isImpThL = ~bmImLineCross;
        end
        if bDid
            isImpDR = isImpDnR;
            isImpDL = isImpDnL;
            isImpUR = isImpThR;
            isImpUL = isImpThL;
        else
            isImpUR = isImpDnR;
            isImpUL = isImpDnL;
            isImpDR = isImpThR;
            isImpDL = isImpThL;
        end
        bUp = xor(bDid, bDense);
        bDown = ~bUp;
        ky_dL = calc_kc(kc_d, kx, 1, isImpDL);
        ky_uL = calc_kc(kc_u, kx, 1, isImpUL);
        ky_dR = calc_kc(kc_d, kx, 1, isImpDR);
        ky_uR = calc_kc(kc_u, kx, 1, isImpUR);
        if bDown
            if isempty(oBcD)
                oBcD_constructor;
                oBcD{2} = oBcD{1};
            end
            oBcD{1} = add_points(oBcD{1}, bv, xiR, jac, kx, ky_dL, ky_uL);
            oBcD{2} = add_points(oBcD{2}, bv, xiR, jac, kx, ky_dR, ky_uR);
        else
            if isempty(oBcU)
                oBcU_constructor;
                oBcU{2} = oBcU{1};
            end
            oBcU{1} = add_points(oBcU{1}, bv, xiR, jac, kx, ky_dL, ky_uL);
            oBcU{2} = add_points(oBcU{2}, bv, xiR, jac, kx, ky_dR, ky_uR);
        end
    end

%%% add_sdp ----------------------------------------------------------------
    function bvFix = add_sdp(bv, bDense)
        nBv = sum(bv);
        bvRefL = bvRef & bv; % a Local version
        if any(bvRefL)
            bvCJL = logical(bvCritJump(bv));
        else
            bvCJL = zeros(nBv, 1);
        end
        if bDense
            vSigma = vSigma_dn(bv);
            vSdp_kx = vSdp_dn(bv);
            vGamma = vGamma_CplxSp(bv);
        else
            vSigma = vSigma_th(bv);
            vSdp_kx = vSdp_th(bv);
            vGamma = pi/4;
        end
        xiR = zeros(nBv, points/2); % These are the actual sampling points
        wR = xiR;                   % Not the ones used to track the path
        if any(bvCJL)
            pow = 0.5;
            [xiR(bvCJL,:), wR(bvCJL,:)] = gausshrhermite(points/2, vSigma(bvCJL), pow);
        end
        if any(~bvCJL)
            pow = 0;
            [xiR(~bvCJL,:), wR(~bvCJL,:)] = gausshrhermite(points/2, vSigma(~bvCJL), pow);
        end
        
        xiR = [-xiR(:,end:-1:1) xiR];
        wR = [wR(:,end:-1:1) wR];
        nTrack = points; % must be even for smooth_path
        TrackLen = 1.001;
        vMin = min(xiR, [], 2)*TrackLen;
        vMax = max(xiR, [], 2)*TrackLen;
        xi = mlinspace(vMin, vMax, nTrack);
        
        g0_l = exp(1j*vGamma);
        g0_r = -g0_l;
        isImpRightTh = zeros(nBv, 1);
        isImpRightDn = (imag(vSdp_kx) + xi(:,end/2+1).*imag(-g0_r))<0;
        isImpLeftTh = (real(vSdp_kx)+xi(:,end/2+1).*real(-g0_l)>0) & (real(vSdp_kx)<kc_th); % when sdp near the origin we start on the proper sheet
        isImpLeftDn = isImpLeftTh | (real(vSdp_kx)>kc_th); % does this work well for cmplx?
        
        xi_sdp_half = xi(:,end/2+1:end);
        mKxSdpLeft = grad_walk(vSdp_kx, g0_l, xi_sdp_half, isImpLeftTh, isImpLeftDn, bv);
        mKxSdpLeft = mKxSdpLeft(:, end:-1:1);
        mKxSdpRight = grad_walk(vSdp_kx, g0_r, xi_sdp_half, isImpRightTh, isImpRightDn, bv);
        oSdp.kx = [mKxSdpLeft mKxSdpRight];
        
        % smooth path
        sm_path = zeros(nBv, points);
        jac = sm_path;
        if any(bvCJL)
            iCJL = find(bvCJL);
            iCJL = iCJL(1);
            x = real(oSdp.kx(bvCJL,:));
            y = imag(oSdp.kx(bvCJL,:));
            vKx0 = repmat(kc_th, sum(bvCJL), 1);
            [sm_path(bvCJL,:), jac(bvCJL,:)] = smooth_path_vect_LS(x, y, xi(iCJL,:), xiR(iCJL,:), vMax(bvCJL), vKx0);
        end
        if any(~bvCJL)
            iCJL = find(~bvCJL);
            iCJL = iCJL(1);
            x = real(oSdp.kx(~bvCJL,:));
            y = imag(oSdp.kx(~bvCJL,:));
            vKx0 = vSdp_kx(~bvCJL);
            [sm_path(~bvCJL,:), jac(~bvCJL,:)] = smooth_path_vect_LS(x, y, xi(iCJL,:), xiR(iCJL,:), vMax(~bvCJL), vKx0);
        end
        
        oSdp.kx = sm_path;
        oSdp.jac = (jac.*wR);
        
        % re-calc isImps
        isImpDn = zeros(nBv, points);
        isImpTh = isImpDn;
        if any(bvCJL)
            isImpDn(bvCJL,:) = (real(oSdp.kx(bvCJL,:)) > 0) & (real(oSdp.kx(bvCJL,:)) < kc_th);
            isImpTh(bvCJL,:) = isImpDn(bvCJL,:);
            oIm.Il(bvCritJump) = 1;
            oIm.y_s(bvCritJump) = -y_s;
            oIm.Er(bvCritJump) = Er_d;
        end
        if any(~bvCJL)
            isImpRightTh = zeros(sum(~bvCJL), 1);
            isImpRightDn = imag(oSdp.kx(~bvCJL, end/2+1))<0;
            isImpLeftTh = (real(oSdp.kx(~bvCJL, end/2)) > 0) & (real(vSdp_kx(~bvCJL)) < kc_th);
            isImpLeftDn = isImpLeftTh | (real(vSdp_kx(~bvCJL)) > kc_th);
            isImpDn(~bvCJL,:) = calc_isImp(oSdp.kx(~bvCJL,:), isImpRightDn, isImpLeftDn, kc_dn);
            isImpTh(~bvCJL,:) = calc_isImp(oSdp.kx(~bvCJL,:), isImpRightTh, isImpLeftTh, kc_th);
        end
        
        % fix possible error in dense bp interception due to inaccurate path.
        % it is possible that the path will not intcp, when the exact path
        % should intcp. This is detected by two conditions:
        % 1. That the exact path intcps
        % 2. That the path ended on the proper sheet.
        if ~bDense
            bvExactIntcp = vFalse_all;
            bvExactIntcp(bvTran) = bvSpDnIntcp;
            bvExactIntcp(bvRef) = bvUpBpIntcp;
            bvExactIntcp = bvExactIntcp(bv);
            bvTwoConditions = ~isImpDn(:,end) & ~isImpTh(:,end) & bvExactIntcp;
            bvFix = vFalse_all;
            bvFix(bv) = bvTwoConditions;
        else
            bvFix = vFalse_all;
        end
        if bDid
            oSdp.ky_d = calc_kc(kc_d, oSdp.kx, 1, isImpDn);
            oSdp.ky_u = calc_kc(kc_u, oSdp.kx, 1, isImpTh);
        else
            oSdp.ky_d = calc_kc(kc_d, oSdp.kx, 1, isImpTh);
            oSdp.ky_u = calc_kc(kc_u, oSdp.kx, 1, isImpDn);
        end
        oSdp.bv = bv;
        if bDense
            if isempty(oSdpDn)
                oSdpDn_constructor;
            end
            oSdpDn = add_points(oSdpDn, oSdp.bv, xiR, oSdp.jac, oSdp.kx,...
                oSdp.ky_d, oSdp.ky_u);
        else
            if isempty(oSdpTh)
                oSdpTh_constructor;
            end
            oSdpTh = add_points(oSdpTh, oSdp.bv, xiR, oSdp.jac, oSdp.kx,...
                oSdp.ky_d, oSdp.ky_u);
        end
    end
    function mKx = grad_walk(kx0, g0, xi, isImpTh, isImpDn, bv)
        nXi = size(xi, 2);
        nBv = sum(bv);
        mKx = zeros(nBv, nXi);
        delta0 = xi(:,1);
        delta = diff(xi, 1, 2);
        delta(:,end+1) = delta(:,end);
        step = -g0.*delta0;
        kx = kx0 + step;
        for n = 1:nXi
            mKx(:,n) = kx;
            dPh = phase_factor_diff(kx, isImpTh, isImpDn, bv);
            g = imag(dPh)+1j*real(dPh);
            next_step = -delta(:,n).*g./abs(g);
            step = next_step;
            kx = kx + step;
            isImpTh = xor(isImpTh, bc_crossed(kx-step, kx, kc_th));
            isImpDn = xor(isImpDn, bc_crossed(kx-step, kx, kc_dn));
        end
    end

    function dPh = phase_factor_diff(s, isImpTh, isImpDn, bv)
        if bDid
            isImpU = isImpTh;
            isImpD = isImpDn;
        else
            isImpD = isImpTh;
            isImpU = isImpDn;
        end
        dPh = vA(bv) + vB(bv).*ky_diff(kc_u, s, isImpU) + vC(bv).*ky_diff(kc_d, s, isImpD);
    end

    function out = ky_diff(kc, kx, isImp)
        out = -kx./calc_kc(kc, kx, 1, isImp);
    end

    function isImp = calc_isImp(mKx, vIsImpR, vIsImpL, kc)
        n2 = size(mKx,2)/2;
        n1 = size(mKx,1);
        mIsImpR = zeros(n1,n2);
        mIsImpL = mIsImpR;
        mIsImpR(:,1) = vIsImpR;
        mIsImpL(:,end) = vIsImpL;
        for iKx = 2:n2
            kxR = mKx(:,iKx+end/2);
            kxL = mKx(:,-iKx+1+end/2);
            kxR_prev = mKx(:,iKx+end/2-1);
            kxL_prev = mKx(:,-iKx+2+end/2);
            vIsImpR = xor(vIsImpR, bc_crossed(kxR_prev, kxR, kc));
            vIsImpL = xor(vIsImpL, bc_crossed(kxL_prev, kxL, kc));
            mIsImpR(:,iKx) = vIsImpR;
            mIsImpL(:,end-iKx+1) = vIsImpL;
        end
        isImp = [mIsImpL mIsImpR];
    end

    function out = bc_crossed(kx1, kx2, kc)
        out = (sign(real(kx1)) ~= sign(real(kx2))) | ...
            ((sign(imag(kx1)) ~= sign(imag(kx2))) & (kc > (real(kx1)-imag(kx1)./imag(kx1-kx2).*real(kx1-kx2))));
    end

%% add_qs -----------------------------------------------------------------
    function add_qs
        % set-up images
        if isMSource
            % Reflected
            oIm.Il(bvRef & bvQs) = (Er_u-Er_d)/(Er_u+Er_d);
            oIm.y_s(bvRef & bvQs) = -y_s;
            oIm.Er(bvRef & bvQs) = Er_d;
            % Transmitted
            oIm.Il(bvTran & bvQs) = 2*Er_u/(Er_u+Er_d);
            oIm.y_s(bvTran & bvQs) = y_s;
            oIm.Er(bvTran & bvQs) = Er_u;
        else
            % Only Trans. Ref has no image
            oIm.Il(bvTran & bvQs) = 1;
            oIm.y_s(bvTran & bvQs) = y_s;
            oIm.Er(bvTran & bvQs) = Er_u;
        end
        oIm.bQs(bvQs) = 1;
        oIm.x_s = repmat(x_s, nX, 1);
        % calc_path
        minX = -pi/2*(1-0.00001);
        maxX = pi/2*(1-0.00001);
        bvRQ = bvRef & bvQs;
        bvTQ = bvTran & bvQs;
        bvRX = bvRQ & bvXgrY;
        bvTX = bvTQ & bvXgrY;
        bvRY = bvRQ & ~bvXgrY;
        bvTY = bvTQ & ~bvXgrY;
        if isempty(oSdpTh)
            oSdpTh_constructor;
        end
        rule_and_path(bvRX, 1);
        rule_and_path(bvRY, 1);
        rule_and_path(bvTX, 0);
        rule_and_path(bvTY, 0);
        oSdpTh.ky_d(bvQs,:) = calc_kc(kc_d, oSdpTh.kx(bvQs,:), 1, 0);
        oSdpTh.ky_u(bvQs,:) = calc_kc(kc_u, oSdpTh.kx(bvQs,:), 1, 0);
        oSdpTh.bv = oSdpTh.bv | bvQs;
        
        function w = weight_fun(t, bR)
            vKx = path(t);
            oPath.kx = vKx;
            oPath.ky_d = calc_kc(kc_d, vKx, 1, 0);
            oPath.ky_u = calc_kc(kc_u, vKx, 1, 0);
            % Only the sign of Y matters for slowly varying:
            Y = -1*bR + 1*~bR;
            slvar = slowly_varying_qs_vect(oPath, x_s+1, Y, x_s,...
                y_s, k0, Er_d, Er_u, sComp, isMSource);
            w = abs(slvar);
        end
        
        function [vKx, vJac] = path(xi)
            [rows, cols] = size(xi);
            xi = xi(:);
            x1 = 2*kc_dn;
            x0 = 0;
            dilate = 0.5;
            A = 1/(x0-x1);
            t = (x1-x0)*dilate*tan(xi)+(x1+x0)/2*0;
            vKx = t+1j*A*(t-x1).*(t-x0);
            vJac = (1+1j*A*(t-x0)+1j*A*(t-x1)).*(x1-x0).*dilate.*(1+tan(xi).^2);
            vJac = reshape(vJac, rows, cols);
            vKx = reshape(vKx, rows, cols);
        end
        function rule_and_path(bv, bR)
            if any(bv)
                switch lower(qsRule)
                    case 'discretization'
                        [xi, w] = gaussGen(points, minX, maxX, @(t)weight_fun(t, bR));
                    case 'legendre'
                        [xi,w] = gaussquad(points, minX, maxX);
                    otherwise
                        error('NDSDP: unrecognized quasi-static rule. Options.quasistaticRule should be either ''discretization'', or ''Legendre''')
                end
                xi = xi(:)'; w = w(:)';
                nBv = sum(bv);
                oSdpTh.xi(bv,:) = makematR(xi, nBv);
                [kx, jac] = path(xi);
                oSdpTh.kx(bv,:) = makematR(kx, nBv);
                jacw = jac.*w;
                oSdpTh.jac(bv,:) = makematR(jacw, nBv);
            end
        end
    end

%% add_sc --------------------------------------------------
    function add_sc
        nSc = sum(bvSc);
        minX = 0;
        maxX = pi;
        factor_sl = 4; %num of points factor
        factor_circ = factor_sl/(factor_sl-2); % to keep total to points
        factor_sl = factor_sl/sc_factor;
        factor_circ = factor_circ/sc_factor;
        [xiC, wC] = gaussquad(round(points/factor_circ), minX, maxX);
        
        xiC = makematR(xiC(:).', nSc);
        wC = makematR(wC(:).', nSc);
        
        % make a semi-circle
        k_tag = (kc_dn+kc_th)/2;
        vK_rad = max((kc_dn-kc_th), min(2*pi./vA(bvSc)/2, k_tag));
        k_rad = makematC(vK_rad, round(points/factor_circ));
        kx_circ = k_tag + exp(1j*(pi-xiC)).*k_rad;
        jac_circ = -1j*k_rad.*exp(1j*(pi-xiC)).*wC;
        
        % make two straight lines
        [xiR, wR] = gausslag(round(points/factor_sl), abs(vAlpha_sdp(bvSc)),0);
        mStep_right_sdp = makematC(vStep_right_sdp(bvSc), round(points/factor_sl));
        k_rad = makematC(vK_rad, round(points/factor_sl));
        kx_right = mStep_right_sdp.*xiR+k_tag+k_rad;
        jac_right = mStep_right_sdp.*wR;
        
        kx_left = -conj(mStep_right_sdp).*xiR(:,end:-1:1)+k_tag-k_rad;
        jac_left = conj(mStep_right_sdp).*wR(:,end:-1:1);
        
        xi = [xiR(:,end:-1:1) xiC xiR];
        kx = [kx_left kx_circ kx_right];
        if isempty(oSc)
            oSc_constructor;
        end
        oSc.xi(bvSc,:) = xi;
        oSc.kx(bvSc,:) = kx;
        oSc.jac(bvSc,:) = [jac_left jac_circ jac_right];
        isImpUD = (real(kx)>0) & (real(kx)<kc_th) & (imag(kx)<0);
        oSc.ky_d(bvSc,:) = calc_kc(kc_d, kx, 1, isImpUD);
        oSc.ky_u(bvSc,:) = calc_kc(kc_u, kx, 1, isImpUD);
        oSc.bv = oSc.bv | bvSc;
    end

%% Utilities -------------------------------------------------------------

    function oOut = squeeze_zeros(oPath)
        bv = oPath.bv;
        oOut = oPath;
        oOut.xi = oPath.xi(bv,:);
        oOut.kx =  oPath.kx(bv,:);
        oOut.jac = oPath.jac(bv,:);
        oOut.ky_d = oPath.ky_d(bv,:);
        oOut.ky_u = oPath.ky_u(bv,:);
    end

    function out = makematR(vIn, n)
        out = repmat(vIn, n, 1);
    end
    function out = makematC(vIn, n)
        out = repmat(vIn, 1, n);
    end

    function alpha = calc_alpha(bDense, step)
        s = 2*((bDid && bDense) || (~bDid && ~bDense))-1;
        s_d = s;
        s_u(bvRef) = s;
        s_u(bvTran) = -s;
        s_u = s_u(:);
        alpha = vA.*imag(step)+(s_u.*abs(vY) + s_d.*abs(y_s)).*abs(real(step));
    end

    function oOut = add_points(oOut, bv, xi, jac, kx, ky_d, ky_u)
        oOut.bv = oOut.bv | bv;
        oOut.xi(bv,:) = xi;
        oOut.jac(bv,:) = jac;
        oOut.kx(bv,:) = kx;
        oOut.ky_d(bv,:) = ky_d;
        oOut.ky_u(bv,:) = ky_u;
    end

%% Constructors -----------------------------------------------------
    function oSdpDn_constructor
        mZeros = zeros(nX, points);
        oSdpDn.xi = mZeros;
        oSdpDn.kx =  mZeros;
        oSdpDn.jac = mZeros;
        oSdpDn.ky_u = mZeros;
        oSdpDn.ky_d = mZeros;
        oSdpDn.bv = zeros(nX, 1);
    end

    function oSc_constructor
        mZeros = zeros(nX, points*sc_factor);
        oSc.xi = mZeros;
        oSc.kx =  mZeros;
        oSc.jac = mZeros;
        oSc.ky_u = mZeros;
        oSc.ky_d = mZeros;
        oSc.bv = zeros(nX, 1);
    end

    function oSdpTh_constructor
        mZeros = zeros(nX, points);
        oSdpTh.xi = mZeros;
        oSdpTh.kx =  mZeros;
        oSdpTh.jac = mZeros;
        oSdpTh.ky_u = mZeros;
        oSdpTh.ky_d = mZeros;
        oSdpTh.bv = zeros(nX, 1);
    end

    function oBcD_constructor
        mZeros = zeros(nX, bc_points);
        oBcD{1}.xi = mZeros;
        oBcD{1}.kx =  mZeros;
        oBcD{1}.jac = mZeros;
        oBcD{1}.ky_u = mZeros;
        oBcD{1}.ky_d = mZeros;
        oBcD{1}.bv = zeros(nX, 1);
    end

    function oBcU_constructor
        mZeros = zeros(nX, bc_points);
        oBcU{1}.xi = mZeros;
        oBcU{1}.kx =  mZeros;
        oBcU{1}.jac = mZeros;
        oBcU{1}.ky_u = mZeros;
        oBcU{1}.ky_d = mZeros;
        oBcU{1}.bv = zeros(nX, 1);
    end
    function ph = phase_factor(s, isImpTh, isImpDn, bv)
        if bDid
            isImpU = isImpTh;
            isImpD = isImpDn;
        else
            isImpD = isImpTh;
            isImpU = isImpDn;
        end
        if any(vB(bv))
            kyup = calc_kc(kc_u, s, 1, isImpU);
        else
            kyup = 0;
        end
        ph = vA(bv).*s + vB(bv).*kyup + vC(bv).*calc_kc(kc_d, s, 1, isImpD);
    end
end