% This function is only used to generate the gaussian rule
% The images are therefore not exacly as in the actual slowly_varying,
% because this would induce a dependence on observation point in the
% gaussian rule generation.
function field = slowly_varying_qs_vect(oPath, vX, vY, x_s, y_s, k0, Er_d, Er_u, sComp, bMSource)
global ETA
mKx = oPath.kx.';
mKy_inc = oPath.ky_d.';
mKy_t = oPath.ky_u.';
beta = 0;
vX = vX(:);
vY = vY(:);
nXY = length(vX);
nKx = numel(mKx);
comp = comp_num(sComp);
k_u = k0*sqrt(Er_u);
k_d = k0*sqrt(Er_d);
ETA_d = ETA/sqrt(Er_d);
ETA_u = ETA/sqrt(Er_u);

mX = repmat(vX, 1, nKx);
mSx = sign(mX-x_s);

cos_i = mKy_inc/k_d;
cos_t = mKy_t/k_u;
denom_para = k_d/k_u*cos_t+cos_i;
denom_para(denom_para==0) = eps;
T_para = 2*cos_i./denom_para;
R_para = -(k_d/k_u*cos_t - cos_i)./denom_para;
denom_perp = k_u/k_d*cos_t+cos_i;
denom_perp(denom_perp==0) = eps;
T_perp = 2*cos_i./denom_perp;
R_perp = T_perp-1;

% subtract images for quasi-static case
T_perp = T_perp - 1;
R_para = R_para - (Er_u-Er_d)/(Er_u + Er_d);
T_para = T_para - 2*Er_u/(Er_u + Er_d);

aK_inc(1,:,:) = mSx.*mKx;
aK_inc(2,:,:) = mKy_inc;
aK_inc(3,:,:) = beta;
aK_r = aK_inc;
aK_r(2,:,:) = -mKy_inc;
aK_t = aK_inc;
aK_t(2,:,:) = mKy_t;
unitY = zeros(size(aK_t));
unitZ = unitY;
unitY(2,:,:) = 1;
unitZ(3,:,:) = 1;
if ~bMSource
    aE0(1,:,:) = -mSx.*beta.*mKx./mKy_inc;
    aE0(2,:,:) = -beta;
    aE0(3,:,:) = (k_d^2-beta^2)./mKy_inc;
    aE0 = -1/4/pi/k_d*ETA_d*aE0;
    aH0 = 1/ETA_d/k_d*cross(aK_inc, aE0);
else
    aH0(1,:,:) = -mSx.*beta.*mKx./mKy_inc;
    aH0(2,:,:) = -beta;
    aH0(3,:,:) = (k_d^2-beta^2)./mKy_inc;
    aH0 = -1/4/pi/k_d/ETA_d*aH0;
    aE0 = -ETA_d/k_d*cross(aK_inc, aH0);
end

if beta == 0
    aA = unitZ;
else
    aA = cross(aK_inc, unitY);
end
norm_2_aA = shiftdim(sum(aA.^2),1);


E_perp_i = shiftdim(sum(aE0.*aA),1); %dot product
H_perp_i = shiftdim(sum(aH0.*aA),1);

E_perp_r = R_perp.*E_perp_i;
E_perp_t = T_perp.*E_perp_i;
H_perp_r = R_para.*H_perp_i;
H_perp_t = T_para.*H_perp_i;

bRef = vY<=0;
mRef = repmat(bRef, 1, nKx);

ft1 = zeros(nXY, nKx);
fr1 = ft1;
field = ft1;
cross_t = cross(aK_t/k_u, aA);
cross_t = shiftdim(cross_t(comp,:,:),1);
cross_r = cross(aK_r/k_d, aA);
cross_r = shiftdim(cross_r(comp,:,:),1);
aA = shiftdim(aA(comp,:,:),1);
if lower(sComp(1)) == 'e'
    ft1(~mRef) = (E_perp_t(~mRef).*aA(~mRef) - ...
        ETA_u*H_perp_t(~mRef).*cross_t(~mRef))./norm_2_aA(~mRef);
else
    ft1(~mRef) = (H_perp_t(~mRef).*aA(~mRef) + ...
        1/ETA_u*E_perp_t(~mRef).*cross_t(~mRef))./norm_2_aA(~mRef);
end
field(~mRef) = ft1(~mRef);

if lower(sComp(1)) == 'e'
    fr1(mRef) = (E_perp_r(mRef).*aA(mRef) - ...
        ETA_d*H_perp_r(mRef).*cross_r(mRef))./norm_2_aA(mRef);
    %     fr1(mRef) = R_perp(mRef);
else
    fr1(mRef) = (H_perp_r(mRef).*aA(mRef) + ...
        1/ETA_d*E_perp_r(mRef).*cross_r(mRef))./norm_2_aA(mRef);
    %    fr1(mRef) = (H_perp_r(mRef).*aA(mRef) + ...
    %        1/ETA_d*E_perp_r(mRef).*cross_r(mRef)*0)./norm_2_aA(mRef);
end

field(mRef) = fr1(mRef);

field = reshape(field, nKx, 1);